#  Copyright 2021  Dom Sekotill <dom.sekotill@kodo.org.uk>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Utilities for integration tests
"""

import tempfile
from contextlib import asynccontextmanager
from pathlib import Path
from typing import AsyncGenerator

import anyio

SOCKET_NAME = 'server.sock'


@asynccontextmanager
async def start_server() -> AsyncGenerator[Path, None]:
	"""
	Return an async context manager that runs a wpa_supplicant server for it's duration
	"""
	with tempfile.TemporaryDirectory() as tmpdir_:
		tmpdir = Path(tmpdir_)
		cmd = ['wpa_supplicant', '-g', str(tmpdir / SOCKET_NAME)]
		async with await anyio.open_process(cmd) as proc:
			# wpa_supplicant is pretty fast; one second to wait for startup should be long
			# enough on most systems…?
			await anyio.sleep(1.0)
			try:
				yield tmpdir / SOCKET_NAME
			finally:
				proc.terminate()
